import typing
import collections.abc

from fundi.resolve import resolve
from fundi.types import CallableInfo


def tree(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: (
        collections.abc.MutableMapping[
            typing.Callable[..., typing.Any], collections.abc.Mapping[str, typing.Any]
        ]
        | None
    ) = None,
) -> collections.abc.Mapping[str, typing.Any]:
    """
    Get tree of dependencies of callable.

    :param scope: container with contextual values
    :param info: callable information
    :param cache: tree generation cache
    :return: Tree of dependencies
    """
    if cache is None:
        cache = {}

    values = {}

    for result in resolve(scope, info, cache):
        name = result.parameter.name
        value = result.value

        if not result.resolved:
            dependency = result.dependency
            assert dependency is not None
            value = tree({**scope, "__fundi_parameter__": result.parameter}, dependency, cache)

            if dependency.use_cache:
                cache[dependency.call] = value

        values[name] = value

    return {"call": info.call, "values": values}


def order(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: (
        collections.abc.MutableMapping[
            typing.Callable[..., typing.Any], list[typing.Callable[..., typing.Any]]
        ]
        | None
    ) = None,
) -> list[typing.Callable[..., typing.Any]]:
    """
    Get resolving order of callable dependencies.

    :param info: callable information
    :param scope: container with contextual values
    :param cache: solvation cache
    :return: order of dependencies
    """
    if cache is None:
        cache = {}

    order_: list[typing.Callable[..., typing.Any]] = []

    for result in resolve(scope, info, cache):
        if not result.resolved:
            assert result.dependency is not None

            value = order(scope, result.dependency, cache)
            order_.extend(value)
            order_.append(result.dependency.call)

            if result.dependency.use_cache:
                cache[result.dependency.call] = value

    return order_



import typing as _typing

from .scan import scan
from .from_ import from_
from . import exceptions
from .resolve import resolve
from .debug import tree, order
from .inject import inject, ainject
from .util import injection_trace, is_configured, get_configuration
from .configurable import configurable_dependency, MutableConfigurationWarning
from .types import CallableInfo, TypeResolver, InjectionTrace, R, Parameter, DependencyConfiguration


FromType: _typing.TypeAlias = _typing.Annotated[R, TypeResolver]
"""Tell resolver to resolve parameter's value by its type, not name"""

__all__ = [
    "scan",
    "tree",
    "order",
    "from_",
    "inject",
    "resolve",
    "ainject",
    "Parameter",
    "exceptions",
    "CallableInfo",
    "TypeResolver",
    "is_configured",
    "InjectionTrace",
    "injection_trace",
    "get_configuration",
    "DependencyConfiguration",
    "configurable_dependency",
    "MutableConfigurationWarning",
]



import typing
from typing import overload
from collections.abc import Generator, AsyncGenerator, Awaitable

T = typing.TypeVar("T", bound=type)
R = typing.TypeVar("R")

@overload
def from_(dependency: T, caching: bool = True) -> T: ...
@overload
def from_(
    dependency: typing.Callable[..., Generator[R, None, None]], caching: bool = True
) -> R: ...
@overload
def from_(dependency: typing.Callable[..., AsyncGenerator[R, None]], caching: bool = True) -> R: ...
@overload
def from_(dependency: typing.Callable[..., Awaitable[R]], caching: bool = True) -> R: ...
@overload
def from_(dependency: typing.Callable[..., R], caching: bool = True) -> R: ...



import types
import typing
import collections.abc

from fundi.types import CallableInfo, ParameterResult, Parameter


def resolve_by_dependency(
    param: Parameter,
    cache: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any],
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any],
) -> ParameterResult:
    dependency = param.from_

    assert dependency is not None

    value = override.get(dependency.call)
    if value is not None:
        if isinstance(value, CallableInfo):
            return ParameterResult(
                param, None, typing.cast(CallableInfo[typing.Any], value), resolved=False
            )

        return ParameterResult(param, value, dependency, resolved=True)

    if dependency.use_cache and dependency.call in cache:
        return ParameterResult(param, cache[dependency.call], dependency, resolved=True)

    return ParameterResult(param, None, dependency, resolved=False)


def resolve_by_type(
    scope: collections.abc.Mapping[str, typing.Any], param: Parameter
) -> ParameterResult:
    annotation = param.annotation

    type_options = (annotation,)

    origin = typing.get_origin(annotation)
    if origin is types.UnionType:
        type_options = tuple(t for t in typing.get_args(annotation) if t is not None)
    elif origin is not None:
        type_options = (origin,)

    for value in scope.values():
        if not isinstance(value, type_options):
            continue

        return ParameterResult(param, value, None, resolved=True)

    return ParameterResult(param, None, None, resolved=False)


def resolve(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any],
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> collections.abc.Generator[ParameterResult, None, None]:
    """
    Try to resolve values from cache or scope for callable parameters

    Recommended use case::

        values = {}
        cache = {}
        for result in resolve(scope, info, cache):
            value = result.value
            name = result.parameter_name

            if not result.resolved:
                value = inject(scope, info, stack, cache)
                cache[name] = value

            values[name] = value


    :param scope: container with contextual values
    :param info: callable information
    :param cache: solvation cache(modify it if necessary while resolving)
    :param override: override dependencies
    :return: generator with solvation results
    """
    from fundi.exceptions import ScopeValueNotFoundError

    if override is None:
        override = {}

    for parameter in info.parameters:
        if parameter.from_:
            yield resolve_by_dependency(parameter, cache, override)
            continue

        if parameter.resolve_by_type:
            result = resolve_by_type(scope, parameter)

            if result.resolved:
                yield result
                continue

        elif parameter.name in scope:
            yield ParameterResult(parameter, scope[parameter.name], None, resolved=True)
            continue

        if parameter.has_default:
            yield ParameterResult(parameter, parameter.default, None, resolved=True)
            continue

        raise ScopeValueNotFoundError(parameter.name, info)



import typing

from fundi.scan import scan
from fundi.types import CallableInfo, TypeResolver


def from_(
    dependency: type | typing.Callable[..., typing.Any], caching: bool = True
) -> TypeResolver | CallableInfo[typing.Any]:
    """
    Use callable or type as dependency for parameter of function

    if dependency parameter is callable the ``fundi.scan.scan`` is used

    if dependency parameter is type the ``fundi.types.TypeResolver`` is returned

    :param dependency: function dependency
    :param caching: Whether to use cached result of this callable or not
    :return: callable information
    """
    if isinstance(dependency, type):
        return TypeResolver(dependency)

    return scan(dependency, caching=caching)



import typing
import collections.abc
from contextlib import ExitStack, AsyncExitStack

from fundi.resolve import resolve
from fundi.types import CallableInfo
from fundi.util import call_sync, call_async, add_injection_trace


def inject(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    stack: ExitStack,
    cache: (
        collections.abc.MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None
    ) = None,
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> typing.Any:
    """
    Synchronously inject dependencies into callable.

    :param scope: container with contextual values
    :param info: callable information
    :param stack: exit stack to properly handle generator dependencies
    :param cache: dependency cache
    :param override: override dependencies
    :return: result of callable
    """
    if info.async_:
        raise RuntimeError("Cannot process async functions in synchronous injection")

    if cache is None:
        cache = {}

    values: dict[str, typing.Any] = {}
    try:
        for result in resolve(scope, info, cache, override):
            name = result.parameter.name
            value = result.value

            if not result.resolved:
                dependency = result.dependency
                assert dependency is not None

                value = inject(
                    {**scope, "__fundi_parameter__": result.parameter},
                    dependency,
                    stack,
                    cache,
                    override,
                )

                if dependency.use_cache:
                    cache[dependency.call] = value

            values[name] = value

        return call_sync(stack, info, values)

    except Exception as exc:
        add_injection_trace(exc, info, values)
        raise exc


async def ainject(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    stack: AsyncExitStack,
    cache: (
        collections.abc.MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None
    ) = None,
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> typing.Any:
    """
    Asynchronously inject dependencies into callable.

    :param scope: container with contextual values
    :param info: callable information
    :param stack: exit stack to properly handle generator dependencies
    :param cache: dependency cache
    :param override: override dependencies
    :return: result of callable
    """
    if cache is None:
        cache = {}

    values: dict[str, typing.Any] = {}

    try:
        for result in resolve(scope, info, cache, override):
            name = result.parameter.name
            value = result.value

            if not result.resolved:
                dependency = result.dependency
                assert dependency is not None

                value = await ainject(
                    {**scope, "__fundi_parameter__": result.parameter},
                    dependency,
                    stack,
                    cache,
                    override,
                )

                if dependency.use_cache:
                    cache[dependency.call] = value

            values[name] = value

        if not info.async_:
            return call_sync(stack, info, values)

        return await call_async(stack, info, values)
    except Exception as exc:
        add_injection_trace(exc, info, values)
        raise exc



import functools
import inspect
import warnings

from fundi import configurable_dependency, MutableConfigurationWarning
from fundi.util import get_configuration, is_configured


def test_configurable_dependency_caching():
    @configurable_dependency
    def factory(require_admin: bool = False):
        def checker(): ...

        return checker

    assert factory() is factory()
    assert factory(require_admin=True) is not factory()
    assert factory(require_admin=False) is factory()
    assert factory(require_admin=True) is factory(require_admin=True)
    assert factory(require_admin=False) is factory(require_admin=False)

    dependency = factory()

    assert is_configured(dependency)

    config = get_configuration(dependency)

    origin = inspect.unwrap(factory)
    assert config.configurator.call is origin

    assert config.values == {"require_admin": False}


def test_configurable_dependency_mutable_argument():
    @configurable_dependency
    def factory(permissions: list[str]):
        def checker(): ...

        return checker

    assert factory(("permissions",)) is factory(("permissions",))

    with warnings.catch_warnings(record=True) as got_warnings:
        assert factory(["permissions"]) is not factory(["permissions"])

    assert got_warnings
    assert got_warnings[0].category is MutableConfigurationWarning

    dependency = factory(("permission",))
    assert is_configured(dependency)

    config = get_configuration(dependency)

    origin = inspect.unwrap(factory)
    assert config.configurator.call is origin

    assert config.values == {"permissions": ("permission",)}
